package be.iminds.ilabt.jfed.experimenter_gui.editor.properties;

import be.iminds.ilabt.jfed.experimenter_gui.config.AdvertisementFetchManager;
import be.iminds.ilabt.jfed.experimenter_gui.config.ImageMap;
import be.iminds.ilabt.jfed.experimenter_gui.config.JFedGuiConfig;
import be.iminds.ilabt.jfed.experimenter_gui.config.TestbedSiteInfo;
import be.iminds.ilabt.jfed.experimenter_gui.editor.properties.map.MapDialogFactory;
import be.iminds.ilabt.jfed.experimenter_gui.ui.ComponentInfoCell;
import be.iminds.ilabt.jfed.experimenter_gui.util.BrowserUtil;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.AdvertisementCounts;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Resource;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.ServerGlimpse;
import be.iminds.ilabt.jfed.lowlevel.lib.JFedConnectivityException;
import be.iminds.ilabt.jfed.rspec.model.HardwareType;
import be.iminds.ilabt.jfed.rspec_fx.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.ui.javafx.FXMLUtil;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import be.iminds.ilabt.jfed.util.common.GeniUrn;
import javafx.application.Platform;
import javafx.beans.property.ObjectProperty;
import javafx.beans.property.ReadOnlyBooleanProperty;
import javafx.beans.property.SimpleObjectProperty;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.collections.ObservableList;
import javafx.collections.transformation.FilteredList;
import javafx.event.ActionEvent;
import javafx.event.EventHandler;
import javafx.fxml.FXML;
import javafx.geometry.Insets;
import javafx.scene.control.*;
import javafx.scene.layout.RowConstraints;
import javafx.scene.layout.StackPane;
import javafx.scene.layout.VBox;
import javafx.scene.paint.Color;
import javafx.scene.text.TextAlignment;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.Objects;
import java.util.Optional;

/**
 * User: twalcari
 * Date: 8/29/2014
 * Time: 11:26
 */
public class ComponentManagerPropertiesPane extends StackPane {
    private final static Logger LOG = LoggerFactory.getLogger(ComponentManagerPropertiesPane.class);

    @Nonnull private final JFedGuiConfig config;
    @Nonnull private final AdvertisementFetchManager advertisementFetchManager;
    @Nonnull private final MapDialogFactory mapDialogFactory;
    @Nonnull private final Resource resource;
    @Nonnull private final ObservableValue<TestbedSiteInfo> chosenSite;
    @Nonnull private final BrowserUtil browserUtil;

    @FXML
    private ComboBox<JFedGuiConfig.ComponentInfo> componentIdComboBox;
    @FXML
    private ComboBox<JFedGuiConfig.HardwareTypeInfo> hardwareTypesComboBox;
    @FXML
    private ToggleGroup selectComponentIdToggleGroup;
    @FXML
    private Label chooseNodeLabel;
    @FXML
    private Button refreshAdvertisementButton;
    @FXML
    private RadioButton componentIdNodeRadioButton, anyNodeRadioButton, hardwareTypeNodeRadioButton;
    @FXML
    private ProgressBar advertisementFetchProgressBar;
    @FXML
    private VBox advertisementFetchOverlay;
    @FXML
    private Button nodesVizButton;

    @FXML
    private Label hardwareTypeInfoLabel;
    @FXML
    private Hyperlink hardwareTypeInfoButton;
    @FXML
    private RowConstraints hardwareInfoRowConstraints;
    @FXML
    private Label hardwareTypeAvailableCountLabel;
    @FXML
    private Label hardwareTypeInfoDetailsLabel;
    @FXML
    private CheckBox onlyAvailableNodes;


    private final ObjectProperty<AdvertisementFetchManager.Resources.RefreshAdvertisementTask> activeRefreshTask =
            new SimpleObjectProperty<>();
    private final ObjectProperty<AdvertisementFetchManager.Resources> activeResources = new SimpleObjectProperty<>();

    public ComponentManagerPropertiesPane(@Nonnull final Resource resource,
                                          @Nonnull FXRspecNode node,
                                          @Nonnull ObservableValue<TestbedSiteInfo> chosenSite,
                                          boolean initialAvailableNodesOnly,
                                          @Nonnull JFedGuiConfig config,
                                          @Nonnull AdvertisementFetchManager advertisementFetchManager,
                                          @Nonnull MapDialogFactory mapDialogFactory,
                                          @Nonnull BrowserUtil browserUtil) {
        assert resource != null;
        assert node != null;
        assert chosenSite != null;

        this.resource = resource;
        this.chosenSite = chosenSite;
        this.config = config;
        this.browserUtil = browserUtil;
        this.advertisementFetchManager = advertisementFetchManager;
        this.mapDialogFactory = mapDialogFactory;

        //initialize FXML
        FXMLUtil.injectFXML(this);

        onlyAvailableNodes.setSelected(initialAvailableNodesOnly);

        Server server = resource.getServer();
        assert server != null : "Resource "+resource.getId()+" has no server";

        //componentIdComboBox.setConverter(StringConverters.XXXX);
        componentIdComboBox.setCellFactory(param -> new ComponentInfoCell(resource, config));
        componentIdComboBox.setButtonCell(new ComponentInfoCell(resource, config) {
            {
                nameLabel.setTextFill(Color.BLACK);
                container.setPadding(new Insets(0, 0, 0, 0));
            }
        });
        componentIdComboBox.setPlaceholder(new Label("--placeholder--"));


        JFedGuiConfig.Scope initialScope = initialAvailableNodesOnly ? JFedGuiConfig.Scope.AVAILABLE_ONLY : JFedGuiConfig.Scope.ALL;
        componentIdComboBox.setItems(filterComponentIdItems(config.getComponentInfo(server, initialScope, resource)));

        hardwareTypesComboBox.setCellFactory(listview -> new HardwareTypeListCell(config, server, browserUtil));

        assert Platform.isFxApplicationThread();
        hardwareTypesComboBox.setItems(config.getHardwareTypes(server, initialScope, resource));

        //1. bind refresh overlay to activeRefreshTask
        advertisementFetchOverlay.visibleProperty().unbind();
        advertisementFetchOverlay.visibleProperty().bind(activeRefreshTask.isNotNull());

        activeRefreshTask.addListener((observable, oldValue, newValue) -> {
            if (oldValue != null) {
                advertisementFetchProgressBar.progressProperty().unbind();
            }

            if (newValue != null) {
                advertisementFetchProgressBar.progressProperty().bind(newValue.progressProperty());
            }
        });

        //2. bind activeRefreshTask to activeResources
        activeResources.addListener((observable, oldValue, newValue) -> {
            activeRefreshTask.unbind();
            activeRefreshTask.bind(newValue.currentRefreshAdvertisementTaskProperty());
        });

        //3. bind activeResources
        activeResources.set(determineResourcesToActivate());

        //initialize bindings
        componentIdComboBox.disableProperty().bind(componentIdNodeRadioButton.selectedProperty().not());
        onlyAvailableNodes.disableProperty().bind(anyNodeRadioButton.selectedProperty());

        hardwareTypesComboBox.disableProperty().bind(hardwareTypeNodeRadioButton.selectedProperty().not());

        hardwareTypeInfoLabel.setVisible(false);

        onlyAvailableNodes.selectedProperty().addListener(observable -> {
            JFedGuiConfig.Scope scope = onlyAvailableNodes.isSelected() ?
                    JFedGuiConfig.Scope.AVAILABLE_ONLY :
                    JFedGuiConfig.Scope.ALL;

            LOG.debug("Will show "+scope+" nodes in comboboxes");

            this.activeResources.setValue(determineResourcesToActivate());
//            if (onlyAvailableNodes.isSelected()) {
//                this.activeResources.setValue(advertisementFetchManager.getAvailableResources(server));
//            } else {
//                this.activeResources.setValue(advertisementFetchManager.getAllResources(server));
//            }

            if (!config.hasUpToDateComponentInfo(server, scope)) {
                refreshAdvertisement();
            }

            assert Platform.isFxApplicationThread();
            componentIdComboBox.setItems(filterComponentIdItems(config.getComponentInfo(server, scope, resource)));
            hardwareTypesComboBox.setItems(config.getHardwareTypes(server, scope, resource));
        });

        chosenSite.addListener((observable, oldValue, newValue) -> {
            JFedGuiConfig.Scope scope = onlyAvailableNodes.isSelected() ?
                    JFedGuiConfig.Scope.AVAILABLE_ONLY :
                    JFedGuiConfig.Scope.ALL;
            componentIdComboBox.setItems(filterComponentIdItems(config.getComponentInfo(server, scope, resource)));
        });

        //initialize pane with values from node
        if (node.getComponentId() != null) {
            Optional<JFedGuiConfig.ComponentInfo> componentId = componentIdComboBox.getItems().stream()
                    .filter(compInfo -> Objects.equals(node.getComponentId(), compInfo.getUrn()))
                    .findFirst();

            if (componentId.isPresent()) {
                componentIdComboBox.getSelectionModel().select(componentId.get());
                componentIdComboBox.setValue(componentId.get());
                componentIdNodeRadioButton.setSelected(true);
            } else {
                //add it to the list of known componentIds in JFedGuiConfig
                JFedGuiConfig.ComponentInfo addedInfo = config.addUserComponentInfo(server.getId(), node.getComponentId());
                componentIdComboBox.getSelectionModel().select(addedInfo);
                componentIdComboBox.setValue(addedInfo);
                componentIdNodeRadioButton.setSelected(true);
            }
        } else if (!node.getHardwareTypes().isEmpty()) {
            if (node.getHardwareTypes().size() > 1) {
                LOG.warn("A node can only have one hardware type in request, but we found {}. Taking first one, and ignoring the rest!", node.getHardwareTypes().size());
            }

            HardwareType userSpecifiedHardwareType = node.getHardwareTypes().get(0);

            JFedGuiConfig.HardwareTypeInfo addedHardwareTypeInfo = config.addUserHardwareTypeInfo(server.getId(), userSpecifiedHardwareType.getName(), null);
            hardwareTypesComboBox.getSelectionModel().select(addedHardwareTypeInfo);
            hardwareTypeNodeRadioButton.setSelected(true);
            componentIdComboBox.getSelectionModel().clearSelection();
            componentIdComboBox.setValue(null);
        } else {
            componentIdComboBox.getSelectionModel().clearSelection();
            componentIdComboBox.setValue(null);
        }

        //add this listener AFTER  the possible addition to hardwareTypesComboBox. This to prevent an unnecessary
        //advertisement call when componentId or hardwareType selected
        //TODO: fedmon-v3: check first if this is cached/known, so that advertisement is not requested when not needed
        //                 probably need to change gui to let user know data is cached, and can be refreshed to get latest
        componentIdNodeRadioButton.selectedProperty().addListener(observable -> {
            if (componentIdNodeRadioButton.isSelected() &&
                    !activeResources.get().isComponentIdsAvailable()) {
                refreshAdvertisement();
            }
        });

        hardwareTypeNodeRadioButton.selectedProperty().addListener(observable -> {
            if (hardwareTypeNodeRadioButton.isSelected() &&
                    !activeResources.get().isComponentIdsAvailable()) {
                refreshAdvertisement();
            }
        });

        //but make sure that a refresh is done if a fixed node assignment is mandatory!
        if (this.resource.getRequireFixedNodeAssignment()) {
            anyNodeRadioButton.setDisable(true);
            componentIdNodeRadioButton.setSelected(true);
        }

        if (!this.resource.getAllowFixedNodeAssignment()) {
            assert !this.resource.getRequireFixedNodeAssignment();
            componentIdNodeRadioButton.setVisible(false);
            componentIdComboBox.setVisible(false);
            nodesVizButton.setVisible(false);

            componentIdNodeRadioButton.setSelected(false);
        }

        if (!this.resource.getAllowFixedNodeAssignment() && !this.resource.getAllowHardwareType()) {
            //if both are not allowed, remove everything related to node selection

            assert !this.resource.getRequireFixedNodeAssignment();
            assert !this.resource.getRequireHardwareType();
            anyNodeRadioButton.setVisible(false);
            chooseNodeLabel.setVisible(false);
            onlyAvailableNodes.setVisible(false);
        }

        if (!this.resource.getAllowFixedNodeAssignment() &&
                !this.resource.getAllowHardwareType() &&
                !this.resource.getAllowDiskImage()) {
            //if these 3 are not allowed, remove everything related to advertisement RSpec
            refreshAdvertisementButton.setVisible(false);
        }

        hardwareTypesComboBox.setVisible(this.resource.getAllowHardwareType());
        hardwareTypeNodeRadioButton.setVisible(this.resource.getAllowHardwareType());

        if (this.resource.getRequireHardwareType()) {
            hardwareTypeNodeRadioButton.setSelected(true); //not really wat we want?
        }

        hardwareTypesComboBox.getSelectionModel().selectedItemProperty().addListener(
                (ob, oldVal, newVal) -> {
                    if (newVal == null) {
                        hardwareTypeAvailableCountLabel.setVisible(false);
                        hardwareTypeInfoLabel.setVisible(false);
                        hardwareInfoRowConstraints.setPrefHeight(0.0);
                    } else {
                        if (newVal.getHardwareTypeName() == null ||
                                config.getHardwareTypeInfo(server).get() == null ||
                                config.getHardwareTypeInfo(server).get().getEntries() == null ||
                                config.getHardwareTypeInfo(server).get().getEntries().isEmpty()) {
                            hardwareTypeInfoLabel.setVisible(false);
                            hardwareInfoRowConstraints.setPrefHeight(0.0);
                        } else {
                            Optional<ServerGlimpse.AdvertisementInfo.HardwareTypeInfo.HardwareTypeInfoEntry> hwTypeInfo =
                                    config.getHardwareTypeInfo(server).get()
                                            .getEntries().stream()
                                            .filter(e -> Objects.equals(e.getName(), newVal.getHardwareTypeName()))
                                            .findAny();
                            hardwareTypeInfoLabel.setVisible(hwTypeInfo.isPresent());
                            if (hwTypeInfo.isPresent()) {
//                            hardwareTypeInfoLabel.setText(newVal.getHardwareTypeName()+" - "+hwTypeInfo.get().getHrn());
                                hardwareTypeInfoLabel.setText(hwTypeInfo.get().getHrn());
                                if (hwTypeInfo.get().getTextDescription() != null) {
//                                hardwareTypeInfoLabel.setTooltip(new Tooltip(hwTypeInfo.get().getTextDescription()));
                                    hardwareTypeInfoDetailsLabel.setText(hwTypeInfo.get().getTextDescription());
                                    hardwareTypeInfoDetailsLabel.setVisible(true);
                                } else {
                                    hardwareTypeInfoDetailsLabel.setText("");
                                    hardwareTypeInfoDetailsLabel.setVisible(false);
                                }
                                if (hwTypeInfo.get().getHtmlDescription() != null) {
                                    hardwareTypeInfoButton.setVisible(true);
                                    hardwareTypeInfoButton.setOnAction(event -> this.browserUtil.openUrlInBrowser(hwTypeInfo.get().getHtmlDescription()));
                                } else {
                                    hardwareTypeInfoButton.setVisible(false);
                                }
                                hardwareInfoRowConstraints.setPrefHeight(USE_COMPUTED_SIZE);
                            } else {
                                hardwareInfoRowConstraints.setPrefHeight(0.0);
                            }
                        }
                        hardwareTypeInfoLabel.getParent().requestLayout();

                        AdvertisementCounts counts = config.getServerNodeCounts(server).get();
                        if (newVal.getHardwareTypeName() == null || counts == null) {
                            hardwareTypeAvailableCountLabel.setVisible(false);
                        } else {
                            String hwType = newVal.getHardwareTypeName();
                            Integer allCount = counts.getHwCount(hwType, AdvertisementCounts.CountScope.ALL);
                            Integer availCount = counts.getHwCount(hwType, AdvertisementCounts.CountScope.AVAILABLE);
                            hardwareTypeAvailableCountLabel.setText(String.format("%d/%d", availCount, allCount));
                            hardwareTypeAvailableCountLabel.setVisible(true);
                        }
                    }
                });

        //check if viz-button must be enabled for this testbed
        if (this.resource.getAllowFixedNodeAssignment()) {
            JFedGuiConfig.NodesVisualization nodesVisualization = config.getNodesVisualization(this.resource.getServer());
            if (nodesVisualization != null) {
                nodesVizButton.disableProperty().bind(componentIdComboBox.disableProperty());

                if (nodesVisualization == JFedGuiConfig.NodesVisualization.MAP) {
                    nodesVizButton.setOnAction(this::showNodesMap);

                    //only show maps-button if running on a 64-bit JVM (see FEDIBBTDEV-1344)
                    String arch = System.getProperty("os.arch");

                    if (Objects.equals(arch, "x86")) {
                        nodesVizButton.setVisible(false);
                    }
                } else if (nodesVisualization == JFedGuiConfig.NodesVisualization.IMAGE) {
                    nodesVizButton.setOnAction(this::showNodesImageMap);
                }
            }
        }

    }

    private ObservableList<JFedGuiConfig.ComponentInfo> filterComponentIdItems(ObservableList<JFedGuiConfig.ComponentInfo> orig) {
        if (chosenSite.getValue() == null) {
            return orig;
        }
        if (!resource.getServer().hasFlag(Server.Flag.featureSiteFilterNodesBySite)) {
            return orig;
        }
        String subAuthName = chosenSite.getValue().getUrn().getSubAuthName(0);
        return new FilteredList<>(orig, ci -> ci.getUrn() == null ||
                (!ci.getUrn().hasSubAuthority()) ||
                ci.getUrn().getSubAuthName(0) == null ||
                ci.getUrn().getSubAuthName(0).equals(subAuthName));
    }

    private AdvertisementFetchManager.Resources determineResourcesToActivate() {
        Server server = resource.getServer();
        assert server != null;

        boolean showAvailableOnly = onlyAvailableNodes.isSelected();

        //if server does not have issues with the "all" advertisement RSpec, always fetch that one.
        //reason: usually, there's no performance or correctness difference between both,
        //        but the all adv rspec does contain more info.
        if (!server.hasFlag(Server.Flag.workaroundRequestAvailableAdvertisement))
            showAvailableOnly = false;

        if (showAvailableOnly)
            return this.advertisementFetchManager.getAvailableResources(server);
        else
            return this.advertisementFetchManager.getAllResources(server);
    }

    private void showNodesImageMap(ActionEvent actionEvent) {
        ImageMap<GeniUrn> imageMap = config.getImageMap(resource.getServer());
        assert imageMap != null;
        JFedGuiConfig.ComponentInfo selectedNode =
                mapDialogFactory.showNodeImageMapDialog(this.getScene().getWindow(), imageMap,
                        config.getComponentInfo(
                                resource.getServer(),
                                JFedGuiConfig.Scope.getScopeForAvailable(onlyAvailableNodes.isSelected()),
                                resource
                        ));

        if (selectedNode != null) {
            componentIdComboBox.setValue(selectedNode);
            componentIdComboBox.getSelectionModel().select(selectedNode);
        }
    }

    private void showNodesMap(ActionEvent actionEvent) {
        JFedGuiConfig.ComponentInfo selectedNode =
                mapDialogFactory.showNodeMapDialog(this.getScene().getWindow(),
                        config.getComponentInfo(
                                resource.getServer(),
                                JFedGuiConfig.Scope.getScopeForAvailable(onlyAvailableNodes.isSelected()),
                                resource)
                );

        if (selectedNode != null) {
            componentIdComboBox.setValue(selectedNode);
            componentIdComboBox.getSelectionModel().select(selectedNode);
        }
    }


    private static boolean isConnectivityException(Throwable t) {

        while (t != null) {
            if (t.getCause() instanceof JFedConnectivityException)
                return true;
            else
                t = t.getCause();
        }
        return false;
    }


    private void refreshAdvertisement() {
        final AdvertisementFetchManager.Resources.RefreshAdvertisementTask updateAdvTask =
                activeResources.get().refreshAdvertisement();

        updateAdvTask.setOnFailed(workerStateEvent -> {
            //only show messages if pane is still visible
            if (ComponentManagerPropertiesPane.this.getParent() != null) {
                if (isConnectivityException(updateAdvTask.getException())) {
                    JFDialogs.create()
                            .owner(ComponentManagerPropertiesPane.this)
                            .message(String.format("A connectivity exception occurred while fetching the advertisement RSpec from %s.\n\n" +
                                            "Verify your internet connection, and consider enabling a proxy in the jFed Preferences.",
                                    resource.getServer().getName()))
                            .masthead("Fetching node advertisement failed")
                            .title("Fetching node advertisement failed")
                            .showException(updateAdvTask.getException());


                } else {
                    JFDialogs.create()
                            .owner(ComponentManagerPropertiesPane.this)
                            .message(String.format("An error occurred while fetching the advertisement RSpec from %s.",
                                    resource.getServer().getName()))
                            .masthead("Fetching advertisement RSpec failed")
                            .title("Fetching advertisement RSpec failed")
                            .showException(updateAdvTask.getException());

                }
            }
        });
    }

    public boolean validate() {
        if (componentIdNodeRadioButton.isSelected() && componentIdComboBox.getSelectionModel().isEmpty()) {
            JFDialogs.create()
                    .owner(this)
                    .message("Please select a node from the list")
                    .masthead("No node selected")
                    .showError();
            return false;
        }

        return true;
    }

    public GeniUrn getSelectedComponentId() {
        if (!componentIdNodeRadioButton.isSelected())
            return null;

        JFedGuiConfig.ComponentInfo res = componentIdComboBox.getSelectionModel().getSelectedItem();
        return res == null ? null : res.getUrn();
    }

    public ReadOnlyBooleanProperty isSpecificNodeChosen() {
        return componentIdNodeRadioButton.selectedProperty();
    }

    public HardwareType getSelectedHardwareType() {
        //special case: forced hardwaretype
        if (resource.getRequireHardwareType() && !resource.getAllowHardwareType() && resource.getHardwareType() != null)
            return new HardwareType(resource.getHardwareType());

        if (!hardwareTypeNodeRadioButton.isSelected())
            return null;

        JFedGuiConfig.HardwareTypeInfo hardwareTypeInfo = hardwareTypesComboBox.getSelectionModel().getSelectedItem();
        String hardwareTypeName = hardwareTypeInfo == null ? null : hardwareTypeInfo.getHardwareTypeName();
        if (hardwareTypeName == null)
            return null;

        return new HardwareType(hardwareTypeName);
    }

    @FXML
    private void onRefreshAdvertisementAction() {
        refreshAdvertisement();
    }

    @FXML
    private void showTestbedMap() {
    }

    public boolean isAvailableNodesOnly() {
        return onlyAvailableNodes.isSelected();
    }
}
